@@ -1,6 +1,6 @@
 /* libpoke.h - Public header for libpoke.  */
 
-/* Copyright (C) 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2020, 2021, 2022, 2023 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
 #ifndef LIBPOKE_H
 #define LIBPOKE_H
 
+#include <stdlib.h>
 #include <stdint.h>
 #include <stdarg.h>
 
@@ -28,8 +29,8 @@
 #define LIBPOKE_API
 #endif
 
-typedef struct pk_compiler *pk_compiler;
-typedef struct pk_ios *pk_ios;
+typedef struct _pk_compiler *pk_compiler;
+typedef struct _pk_ios *pk_ios;
 typedef uint64_t pk_val;
 
 /* The following status codes are returned by pk_errno function.  */
@@ -40,6 +41,29 @@ typedef uint64_t pk_val;
 #define PK_EEOF 3
 #define PK_EINVAL 4
 
+/* The following macros are standard exception codes defined in
+   pkl-rt.pk.  */
+
+#define PK_EC_GENERIC       0
+#define PK_EC_DIV_BY_ZERO   1
+#define PK_EC_NO_IOS        2
+#define PK_EC_NO_RETURN     3
+#define PK_EC_OUT_OF_BOUNDS 4
+#define PK_EC_EOF           6
+#define PK_EC_MAP           7
+#define PK_EC_CONV          8
+#define PK_EC_ELEM          9
+#define PK_EC_CONSTRAINT   10
+#define PK_EC_IO           11
+#define PK_EC_SIGNAL       12
+#define PK_EC_IOFLAGS      13
+#define PK_EC_INVAL        14
+#define PK_EC_EXIT         15
+#define PK_EC_ASSERT       16
+#define PK_EC_OVERFLOW     17
+#define PK_EC_PERM         18
+#define PK_EC_STACK        19
+
 struct pk_color
 {
   int red;
@@ -65,12 +89,12 @@ struct pk_term_if
   void (*indent_fn) (unsigned int lvl, unsigned int step);
 
   /* Mark the beginning of a styling class with name CLASS.  */
-  void (*class_fn) (const char *class);
+  void (*class_fn) (const char *aclass);
 
   /* Mark the end of a styling class with name CLASS.  This function
      returns 0 if the given class is not active and therefore can't be
      ended.  1 otherwise.  */
-  int (*end_class_fn) (const char *class);
+  int (*end_class_fn) (const char *aclass);
 
   /* Mark the beginning of an hyperlink with url URL and identifier
      ID.  The identifier can be NULL.  */
@@ -108,6 +132,14 @@ struct pk_term_if
 
 pk_compiler pk_compiler_new (struct pk_term_if *term_if) LIBPOKE_API;
 
+/* Like pk_compiler_new, but accepts an extra argument with flags,
+   which are defined below.  */
+
+#define PK_F_NOSTDTYPES 1 /* Do not define standard types.  */
+
+pk_compiler pk_compiler_new_with_flags (struct pk_term_if *term_if,
+                                        uint32_t flags) LIBPOKE_API;
+
 /* Destroy an instance of a Poke incremental compiler.
 
    PKC is a previously created incremental compiler.  */
@@ -123,18 +155,19 @@ void pk_compiler_free (pk_compiler pkc) LIBPOKE_API;
 
 int pk_errno (pk_compiler pkc) LIBPOKE_API;
 
-/* Compile a Poke program from the given file FILENAME.
+/* Compile an execute a Poke program from the given file FILENAME.
 
-   If not NULL, *EXIT_STATUS is set to the status resulting from the
-   execution of the program.
+   EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
+   Exception value if the execution rsults in an unhandled exception,
+   PK_NULL otherwise.
 
    Return PK_ERROR in case of a compilation error.  Otherwise,
    return PK_OK.  */
 
 int pk_compile_file (pk_compiler pkc, const char *filename,
-                     int *exit_status) LIBPOKE_API;
+                     pk_val *exit_exception) LIBPOKE_API;
 
-/* Compile a Poke program from a memory buffer.
+/* Compile an execute a Poke program from a memory buffer.
 
    BUFFER is a NULL-terminated string.
 
@@ -142,29 +175,90 @@ int pk_compile_file (pk_compiler pkc, const char *filename,
    not part of the compiled entity.
 
    Return PK_ERROR in case of a compilation error.  Otherwise,
-   return PK_OK.  */
+   return PK_OK.
+
+   EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
+   Exception value if the execution rsults in an unhandled exception,
+   PK_NULL otherwise.  */
 
 int pk_compile_buffer (pk_compiler pkc, const char *buffer,
-                       const char **end) LIBPOKE_API;
+                       const char **end, pk_val *exit_exception) LIBPOKE_API;
+
+/* Like pk_compile_buffer, but gets the following extra arguments.
+
+   SOURCE is a string identifying the source of the code.  This will
+   often be a file name, or a string like "<stdin>".
+
+   LINE and COLUMN specify the relative location of BUFFER in SOURCE.
+   Both lines and columns start numbering at one, i.e. the first
+   column/line is the column/line number 1.  */
+
+int pk_compile_buffer_with_loc (pk_compiler pkc, const char *buffer,
+                                const char *source,
+                                uint32_t line, uint32_t column,
+                                const char **end, pk_val *exit_exception)
+  LIBPOKE_API;
 
 /* Like pk_compile_buffer but compile and execute a single Poke
    statement, which may evaluate to a value if it is an "expression
    statement".
 
    VAL, if given, is a pointer to a pk_val variable that is set to the
-   result value of an expression-statement, or to PK_NULL.  */
+   result value of an expression-statement, or to PK_NULL.
+
+   EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
+   Exception value if the execution rsults in an unhandled exception,
+   PK_NULL otherwise.  */
 
 int pk_compile_statement (pk_compiler pkc, const char *buffer,
-                          const char **end, pk_val *val) LIBPOKE_API;
+                          const char **end, pk_val *val,
+                          pk_val *exit_exception) LIBPOKE_API;
+
+/* Like pk_compile_statement, but gets the following extra arguments.
+
+   SOURCE is a string identifying the source of the code.  This will
+   often be a file name, or a string like "<stdin>".
+
+   LINE and COLUMN specify the relative location of BUFFER in SOURCE.
+   Both lines and columns start numbering at one, i.e. the first
+   column/line is the column/line number 1.  */
+
+int pk_compile_statement_with_loc (pk_compiler pkc, const char *buffer,
+                                   const char *source,
+                                   uint32_t line, uint32_t column,
+                                   const char **end, pk_val *val,
+                                   pk_val *exit_exception) LIBPOKE_API;
 
 /* Like pk_compile_buffer but compile and execute a single Poke
    expression, which evaluates to a value.
 
    VAL, if given, is a pointer to a pk_val variable that is set to the
-   result value of executing the expression.  */
+   result value of executing the expression.
+
+   Return PK_ERROR in case of a compilation error, PK_OK otherwise.
+
+   EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
+   Exception value if the execution rsults in an unhandled exception,
+   PK_NULL otherwise.  */
 
 int pk_compile_expression (pk_compiler pkc, const char *buffer,
-                           const char **end, pk_val *val) LIBPOKE_API;
+                           const char **end, pk_val *val,
+                           pk_val *exit_exception) LIBPOKE_API;
+
+/* Like pk_compile_expression, but gets the following extra arguments.
+
+   SOURCE is a string identifying the source of the code.  This will
+   often be a file name, or a string like "<stdin>".
+
+   LINE and COLUMN specify the relative location of BUFFER in SOURCE.
+   Both lines and columns start numbering at one, i.e. the first
+   column/line is the column/line number 1.  */
+
+int pk_compile_expression_with_loc (pk_compiler pkc, const char *buffer,
+                                    const char *source,
+                                    uint32_t line, uint32_t column,
+                                    const char **end, pk_val *val,
+                                    pk_val *exit_exception) LIBPOKE_API;
 
 /* Load a module using the given compiler.
 
@@ -230,13 +324,48 @@ void pk_set_quiet_p (pk_compiler pkc, int quiet_p) LIBPOKE_API;
 
 /* Install a handler for alien tokens in the incremental compiler.
    The handler gets a string with the token identifier (for $foo it
-   would get `foo') and should return a string containing the
-   resolving identifier.  If the handler returns NULL, then the alien
-   token is not recognized as such in the compiler and ERRMSG points
-   to an explicative error message.  */
+   would get `foo') and should return a pk_alien_token struct with the
+   resulting token.  If the handler returns NULL, then the alien token
+   is not recognized as such in the compiler and ERRMSG points to an
+   explicative error message.  */
 
-typedef char *(*pk_alien_token_handler_fn) (const char *id,
-                                            char **errmsg);
+#define PK_ALIEN_TOKEN_IDENTIFIER 0
+#define PK_ALIEN_TOKEN_INTEGER    1
+#define PK_ALIEN_TOKEN_OFFSET     2
+#define PK_ALIEN_TOKEN_STRING     3
+
+struct pk_alien_token
+{
+  int kind;
+  union
+  {
+    char *identifier;
+
+    struct
+    {
+      uint64_t magnitude;
+      int width;
+      int signed_p;
+    } integer;
+
+    struct
+    {
+      uint64_t magnitude;
+      int width;
+      int signed_p;
+      uint64_t unit;
+    } offset;
+
+    struct
+    {
+      const char *str;
+    } string;
+
+  } value;
+};
+
+typedef struct pk_alien_token *(*pk_alien_token_handler_fn) (const char *id,
+                                                             char **errmsg);
 void pk_set_alien_token_fn (pk_compiler pkc,
                             pk_alien_token_handler_fn cb) LIBPOKE_API;
 
@@ -308,17 +437,24 @@ int pk_ios_get_id (pk_ios ios) LIBPOKE_API;
 
 /* Return the name of the device interface.  */
 
-char *pk_ios_get_dev_if_name (pk_ios ios) LIBPOKE_API;
+const char *pk_ios_get_dev_if_name (pk_ios ios) LIBPOKE_API;
 
-/* Return the size of the given IO space, in bits.  */
+/* Return the size of the given IO space, in bytes.  */
 
 uint64_t pk_ios_size (pk_ios ios) LIBPOKE_API;
 
+/* Return the bias of the given IO space, in bits.
+
+   Each IO space has a bias associated with it, which by default is 0
+   bits.  This bias is applied to the offset given to every read/write
+   operation.  */
+
+uint64_t pk_ios_get_bias (pk_ios ios) LIBPOKE_API;
+
 /* Return the flags which are active in a given IOS.  */
 
 #define PK_IOS_F_READ     1
 #define PK_IOS_F_WRITE    2
-#define PK_IOS_F_TRUNCATE 8
 #define PK_IOS_F_CREATE  16
 
 uint64_t pk_ios_flags (pk_ios ios) LIBPOKE_API;
@@ -361,7 +497,32 @@ void pk_ios_map (pk_compiler pkc, pk_ios_map_fn cb, void *data) LIBPOKE_API;
    determines what subset of declarations will be processed by the
    handler.
 
-   HANDLER is a function of type pk_map_decl_fn.  */
+   HANDLER is a function of type pk_map_decl_fn.  This callback gets
+   the following arguments when invoked by pk_decl_map:
+
+     KIND is one of the PK_DECL_KIND_* values defined below.
+
+     SOURCE is a NULL-terminated string somehow denoting the source of
+     the declaration.  This may be a file name, or a descriptive
+     string like "<stdin>".
+
+     NAME is a NULL-terminated string with the name of the entity
+     that is declared.
+
+     TYPE is a NULL-terminated string with the printed representation
+     of the type of the entity that is declared.  This is NULL for
+     PKL_DECL_KIND_TYPE.
+
+     FIRST_LINE, LAST_LINE, FIRST_COLUMN and LAST_COLUMN contain the
+     location of the declaration of the entity that is declared.
+
+     VALUE is either the value of the declared entity, or PK_NULL for
+     PK_DECL_KIND_TYPE.
+
+     DATA is a user-provided pointer at pk_decl_map invocation.
+
+   DATA is a pointer that is passed to the provided callback.  This
+   can be NULL.  */
 
 #define PK_DECL_KIND_VAR 0
 #define PK_DECL_KIND_FUNC 1
@@ -373,6 +534,7 @@ typedef void (*pk_map_decl_fn) (int kind,
                                 const char *type,
                                 int first_line, int last_line,
                                 int first_column, int last_column,
+                                pk_val value,
                                 void *data);
 void pk_decl_map (pk_compiler pkc, int kind,
                   pk_map_decl_fn handler, void *data) LIBPOKE_API;
@@ -392,7 +554,7 @@ pk_val pk_decl_val (pk_compiler pkc, const char *name) LIBPOKE_API;
 /* Given the name of a variable declared in the compiler, set a new
    value to it.
 
-   If there is no varaible defined with NAME, then this is a
+   If there is no variable defined with NAME, then this is a
    no-operation.  */
 
 void pk_decl_set_val (pk_compiler pkc, const char *name, pk_val val)
@@ -410,17 +572,30 @@ int pk_defvar (pk_compiler pkc, const char *varname, pk_val val) LIBPOKE_API;
 
    CLS is the closure for the function to call.
 
-   RET is set to the value returned by the function, or to PK_NULL if
+   RET is set to the value returned by the function, or to NULL if
    it is a void function.
 
-   A variable number of function arguments follow, terminated by
-   PK_NULL.
+   EXIT_EXCEPTION, if not NULL, is a pointer to a pk_val variable that
+   is set to an Exception value if the function call results in an
+   unhandled exception, PK_NULL otherwise.
+
+   NARG is the number of arguments that follow and that are passed in
+   the function call.  Note that PK_NULL values should be added for
+   "not specified" actuals for formals that have default values.
+
+   NARG function arguments follow.
 
    Return PK_ERROR if there is a problem performing the operation, or if the
    execution of the function results in an unhandled exception.
-   Return PK_OK otherwise.  */
+   Return PK_OK otherwise.
 
-int pk_call (pk_compiler pkc, pk_val cls, pk_val *ret, ...) LIBPOKE_API;
+   Note that calling a method needs for the corresponding struct value
+   holding the method to be passed as the _last_ argument in the
+   pk_call.  */
+
+int pk_call (pk_compiler pkc, pk_val cls,
+             pk_val *ret, pk_val *exit_exception,
+             int narg, ...) LIBPOKE_API;
 
 /* Get and set properties of the incremental compiler.  */
 
@@ -439,6 +614,9 @@ void pk_set_oindent (pk_compiler pkc, unsigned int oindent) LIBPOKE_API;
 int pk_omaps (pk_compiler pkc) LIBPOKE_API;
 void pk_set_omaps (pk_compiler pkc, int omaps_p) LIBPOKE_API;
 
+int pk_autoremap (pk_compiler pkc) LIBPOKE_API;
+void pk_set_autoremap (pk_compiler pkc, int autoremap_p) LIBPOKE_API;
+
 enum pk_omode
   {
     PK_PRINT_FLAT,
@@ -581,6 +759,32 @@ pk_val pk_make_struct (pk_val nfields, pk_val type) LIBPOKE_API;
 
 pk_val pk_struct_nfields (pk_val sct) LIBPOKE_API;
 
+/* Get the value of a struct field given the name of the field.
+
+   SCT is the field value.
+
+   FNAME is a C string with the name of the field in the struct.
+
+   If FNAME doesn't identify a field in the struct, PK_NULL is
+   returned.  */
+
+pk_val pk_struct_ref_field_value (pk_val sct, const char *fname)
+  LIBPOKE_API;
+
+/* Set the value of a struct field given the name of the field.
+
+   SCT is the struct value.
+
+   FNAME is a C string with the name of the field in the struct.
+
+   VALUE is the new value for this field.
+
+   If FNAME doesn't identify a field in the struct, struct remains
+   unchanged.  */
+
+void pk_struct_ref_set_field_value (pk_val sct, const char *fname,
+                                    pk_val value) LIBPOKE_API;
+
 /* Get the bit-offset of the field of a struct, relative to the
    beginning of the struct.
 
@@ -678,7 +882,7 @@ pk_val pk_array_nelem (pk_val array) LIBPOKE_API;
 
    If IDX is invalid, PK_NULL is returned. */
 
-pk_val pk_array_elem_val (pk_val array, uint64_t idx) LIBPOKE_API;
+pk_val pk_array_elem_value (pk_val array, uint64_t idx) LIBPOKE_API;
 
 /* Insert a new element into an array.
 
@@ -718,6 +922,10 @@ void pk_array_set_elem (pk_val array, uint64_t idx, pk_val val) LIBPOKE_API;
 
 pk_val pk_array_elem_boffset (pk_val array, uint64_t idx) LIBPOKE_API;
 
+/* Return the size of VAL, in bits.  */
+
+uint64_t pk_sizeof (pk_val val) LIBPOKE_API;
+
 /* Integral types.  */
 
 /* Build and return an integral type.
@@ -785,7 +993,8 @@ pk_val pk_offset_type_unit (pk_val type) LIBPOKE_API;
 pk_val pk_make_struct_type (pk_val nfields, pk_val name, pk_val *fnames,
                             pk_val *ftypes) LIBPOKE_API;
 
-/* Get the type of a struct.  */
+/* Get the type of a struct.
+   This function is DEPRECATED; please use pk_typeof instead.  */
 
 pk_val pk_struct_type (pk_val sct) LIBPOKE_API;
 
@@ -795,8 +1004,8 @@ void pk_allocate_struct_attrs (pk_val nfields,
 pk_val **fnames, pk_val **ftypes) LIBPOKE_API;
 
 /* Get the name of a struct type.
-
-   If the struct type is anonymous, PK_NULL is returned.  */
+   If the struct type is anonymous, PK_NULL is returned.
+   This function is DEPRECATED; please use pk_type_name instead.  */
 
 pk_val pk_struct_type_name (pk_val type) LIBPOKE_API;
 
@@ -859,10 +1068,12 @@ void pk_struct_type_set_ftype (pk_val type, uint64_t idx,
 /* Build and return an array type.
 
    ETYPE is the type of the elements of the array.
-
-   BOUND is an uint<64> with the number of elements of the
-   array... XXX or a closure?? cant remember. At the moment the PKL
-   compiler generates PVM_NULL for this... */
+   BOUND is either:
+   - A ()any closure returning PK_NULL to denote an unbounded array.
+   - A closure ()uint<64> to denote an array type bounded by number of
+     elements.
+   - A closure ()offset<uint<64>,1> to denote an array type bounded by
+     size.  */
 
 pk_val pk_make_array_type (pk_val etype, pk_val bound) LIBPOKE_API;
 
@@ -870,65 +1081,214 @@ pk_val pk_make_array_type (pk_val etype, pk_val bound) LIBPOKE_API;
 
 pk_val pk_array_type_etype (pk_val type) LIBPOKE_API;
 
-/* Get the bound of the given array type.  */
+/* Get the bound of the given array type.
+   This is always PK_NULL at this point.  */
 
 pk_val pk_array_type_bound (pk_val type) LIBPOKE_API;
 
 /* Mapped values.  */
 
+/* Return a boolean indicating whether the given value is mappable or
+   not.  */
+
+int pk_val_mappable_p (pk_val val) LIBPOKE_API;
+
 /* Return a boolean indicating whether the given value is mapped or
    not.  */
 
 int pk_val_mapped_p (pk_val val) LIBPOKE_API;
 
+/* Set MAPPED_P flag of the given value indicating whether the given
+   value is mapped or not.
+
+   For non-mappable values, this is a no-operation.  Note that this
+   just sets the flag.  */
+
+void pk_val_set_mapped (pk_val val, int mapped_p) LIBPOKE_API;
+
+/* Return a boolean indicating whether data integrity for the given
+   value is enforced or not.  */
+
+int pk_val_strict_p (pk_val val) LIBPOKE_API;
+
+/* Set STRICT flag of the given value indicating whether data
+   integrity is enforced or not.
+
+   For non-mappable values, this is a no-operation.  Note that this
+   just sets the flag.  */
+
+void pk_val_set_strict (pk_val val, int strict_p) LIBPOKE_API;
+
 /* Return the IOS identifier, an int<32>, in which the given value is
    mapped.  If the value is not mapped, return PK_NULL.  */
 
 pk_val pk_val_ios (pk_val val) LIBPOKE_API;
 
-/* Return the offset in which the given value is mapped.
+/* Set the IOS, an int<32>, in which the given value is mapped.
+
+   For non-mappable values, this is a no-operation.  Note that this
+   just sets the IOS property.  */
+
+void pk_val_set_ios (pk_val val, pk_val ios) LIBPOKE_API;
+
+/* Return the offset in which the given value is mapped, as an offset value.
    If the value is not mapped, return PK_NULL.  */
 
 pk_val pk_val_offset (pk_val val) LIBPOKE_API;
 
+/* Set the offset in which the given value is mapped.
+
+   For non-mappable values, this is a no-operation.  Note that this
+   just sets the property.  */
+
+void pk_val_set_offset (pk_val val, pk_val offset) LIBPOKE_API;
+
+/* Return the bit offset in which the given value is mapped, as an
+   uint<64>.
+
+   If the value is not mapped, return PK_NULL.  */
+
+pk_val pk_val_boffset (pk_val val) LIBPOKE_API;
+
+/* Set the bit offset in which the given value is mapped.
+   BOFFSET is expected to be an uint<64>.
+
+   For non-mappable values, this is a no-operation.  Note that this
+   just sets the property.  */
+
+void pk_val_set_boffset (pk_val val, pk_val boffset) LIBPOKE_API;
+
 /* Other operations on values.  */
 
-/* Return the type of the given value.  */
+/* Return the type of the given value.  If the value is itself a type,
+   return it.  */
 
 pk_val pk_typeof (pk_val val) LIBPOKE_API;
 
 /* Given a type value, return its code.  */
 
-#define PK_UNKNOWN 0
-#define PK_INT     1
-#define PK_UINT    2
-#define PK_STRING  3
-#define PK_OFFSET  4
-#define PK_ARRAY   5
-#define PK_STRUCT  6
-#define PK_CLOSURE 7
-#define PK_ANY     8
+#define PK_TYPE_UNKNOWN 0
+#define PK_TYPE_INT     1
+#define PK_TYPE_UINT    2
+#define PK_TYPE_STRING  3
+#define PK_TYPE_OFFSET  4
+#define PK_TYPE_ARRAY   5
+#define PK_TYPE_STRUCT  6
+#define PK_TYPE_CLOSURE 7
+#define PK_TYPE_VOID    8
 
 int pk_type_code (pk_val val) LIBPOKE_API;
 
+/* Given a type value, return its name.
+
+   If the provided type is anonymous, PK_NULL is returned.  */
+
+pk_val pk_type_name (pk_val type) LIBPOKE_API;
+
+/* Given a value, return its kind.  */
+
+#define PK_VAL_UNKNOWN 0
+#define PK_VAL_INT     1
+#define PK_VAL_UINT    2
+#define PK_VAL_STRING  3
+#define PK_VAL_OFFSET  4
+#define PK_VAL_ARRAY   5
+#define PK_VAL_STRUCT  6
+#define PK_VAL_CLOSURE 7
+#define PK_VAL_TYPE    8
+
+int pk_val_kind (pk_val val) LIBPOKE_API;
+
 /* Compare two Poke values.
 
    Returns 1 if they match, 0 otherwise.  */
 
 int pk_val_equal_p (pk_val val1, pk_val val2) LIBPOKE_API;
 
-/* Print the given value using the current PKC settings.  */
+/* Print the given value using the current PKC settings.
 
-void pk_print_val (pk_compiler pkc, pk_val val) LIBPOKE_API;
+   If printing the value implies to run a struct pretty-printer or to
+   execute some other code in the PVM, *EXIT_EXCEPTION is set to an
+   exception value if the operation gets interrupted by an unhandled
+   exception.  Otherwise *EXIT_EXCEPTION is set to PK_NULL.  NULL can
+   be passed for this argument as well.  */
 
-/* Print the given value using the given settings.  */
+void pk_print_val (pk_compiler pkc, pk_val val,
+                   pk_val *exit_exception) LIBPOKE_API;
 
-#define PK_PRINT_F_MAPS   1  /* Output value and element offsets.  */
-#define PK_PRINT_F_PPRINT 2  /* Use pretty-printers.  */
+/* Foreign IO devices interface.  */
 
-void pk_print_val_with_params (pk_compiler pkc, pk_val val,
-                               int depth, int mode, int base,
-                               int indent, int acutoff,
-                               uint32_t flags) LIBPOKE_API;
+typedef uint64_t pk_iod_off;
+
+/* These must match the IOD_* values in ios-dev.h  */
+#define PK_IOD_OK      0
+#define PK_IOD_ERROR  -1
+#define PK_IOD_EIOFF  -2
+#define PK_IOD_EFLAGS -3
+#define PK_IOD_ENOMEM -4
+#define PK_IOD_EOF    -5
+#define PK_IOD_EINVAL -6
+
+struct pk_iod_if
+{
+  /* Return the name of this device interface. Cannot return NULL. */
+
+  const char *(*get_if_name) ();
+
+  /* Determine whether the provided HANDLER is recognized as a valid
+     device spec by this backend, and if so, return its normalized
+     form (caller will free).  In case of error, return NULL.  This function
+     sets the ERROR to error code or to IOD_OK.  */
+  char * (*handler_normalize) (const char *handler, uint64_t flags, int* error);
+
+  /* Open a device using the provided HANDLER.  Return the opened
+     device, or NULL in case of errors.  Set the ERROR to error code or to
+     IOD_OK.  Note that this function assumes that HANDLER is recognized as a
+     handler by the backend.  DATA is the `data' pointer below, intended to be
+     used as an IOD-specific payload specified by the author of IOD.  */
+  void * (*open) (const char *handler, uint64_t flags, int *error, void *data);
+
+  /* Close the given device.  Return the error code if there was an error
+     during the operation, IOD_OK otherwise.  */
+  int (*close) (void *dev);
+
+  /* Read a small byte buffer from the given device at the given byte offset.
+     Return 0 on success, or IOD_EOF on error, including on short reads.  */
+  int (*pread) (void *dev, void *buf, size_t count, pk_iod_off offset);
+
+  /* Write a small byte buffer to the given device at the given byte offset.
+     Return 0 on success, or IOD_EOF on error, including short writes.  */
+  int (*pwrite) (void *dev, const void *buf, size_t count, pk_iod_off offset);
+
+  /* Return the flags of the device, as it was opened.  */
+  uint64_t (*get_flags) (void *dev);
+
+  /* Return the size of the device, in bytes.  */
+  pk_iod_off (*size) (void *dev);
+
+  /* If called on a in-stream, free the buffer before OFFSET.  If called on
+     an out-stream, flush the data till OFFSET and free the buffer before
+     OFFSET.  Otherwise, do not do anything.  Return IOS_OK in success and
+     an error code on failure.  */
+  int (*flush) (void *dev, pk_iod_off offset);
+
+  /* IOD-specific data payload that will be passed to OPEN function.  If not
+     used, it should be NULL.  */
+  void *data;
+};
+
+/* Register a foreign IO device in the Poke compiler.
+
+   IOD_IF is a pointer to a struct pk_iod_if containing pointers to
+   functions providing the IO device implementation.
+
+   At the moment it is only supported to register just one foreign IO
+   device.
+
+   Return PK_ERROR if some error occurs.
+   Return PK_OK otherwise. */
+
+int pk_register_iod (pk_compiler pkc, struct pk_iod_if *iod_if)
+  LIBPOKE_API;
 
 #endif /* ! LIBPOKE_H */
